@@ -27,8 +27,25 @@
 define(["jquery", "footable", "humanize"],
     function ($, _, Humanize) {
         "use strict";
+        var page_size = {
+            errors: 25,
+            history: 25
+        };
+
+        function set_page_size(n, callback) {
+            if (n !== page_size.history && n > 0) {
+                page_size.history = n;
+                if (callback) {
+                    return callback(n);
+                }
+            }
+            return null;
+        }
+
+        set_page_size($("#history_page_size").val());
+
         var ui = {};
-        var ft = {};
+        var prevVersion = null;
         var htmlEscapes = {
             "&": "&amp;",
             "<": "&lt;",
@@ -40,16 +57,17 @@ define(["jquery", "footable", "humanize"],
             "=": "&#x3D;"
         };
         var htmlEscaper = /[&<>"'/`=]/g;
+        var symbols = [];
         var symbolDescriptions = {};
 
-        var EscapeHTML = function (string) {
-            return (String(string)).replace(htmlEscaper, function (match) {
+        var escapeHTML = function (string) {
+            return String(string).replace(htmlEscaper, function (match) {
                 return htmlEscapes[match];
             });
         };
 
         var escape_HTML_array = function (arr) {
-            arr.forEach(function (d, i) { arr[i] = EscapeHTML(d); });
+            arr.forEach(function (d, i) { arr[i] = escapeHTML(d); });
         };
 
         function unix_time_format(tm) {
@@ -59,31 +77,32 @@ define(["jquery", "footable", "humanize"],
 
         function preprocess_item(item) {
             for (var prop in item) {
+                if (!{}.hasOwnProperty.call(item, prop)) continue;
                 switch (prop) {
-                case "rcpt_mime":
-                case "rcpt_smtp":
-                    escape_HTML_array(item[prop]);
-                    break;
-                case "symbols":
-                    Object.keys(item.symbols).map(function (key) {
-                        var sym = item.symbols[key];
-                        if (!sym.name) {
-                            sym.name = key;
-                        }
-                        sym.name = EscapeHTML(sym.name);
-                        if (sym.description) {
-                            sym.description = EscapeHTML(sym.description);
-                        }
+                    case "rcpt_mime":
+                    case "rcpt_smtp":
+                        escape_HTML_array(item[prop]);
+                        break;
+                    case "symbols":
+                        Object.keys(item.symbols).forEach(function (key) {
+                            var sym = item.symbols[key];
+                            if (!sym.name) {
+                                sym.name = key;
+                            }
+                            sym.name = escapeHTML(sym.name);
+                            if (sym.description) {
+                                sym.description = escapeHTML(sym.description);
+                            }
 
-                        if (sym.options) {
-                            escape_HTML_array(sym.options);
+                            if (sym.options) {
+                                escape_HTML_array(sym.options);
+                            }
+                        });
+                        break;
+                    default:
+                        if (typeof item[prop] === "string") {
+                            item[prop] = escapeHTML(item[prop]);
                         }
-                    });
-                    break;
-                default:
-                    if (typeof (item[prop]) === "string") {
-                        item[prop] = EscapeHTML(item[prop]);
-                    }
                 }
             }
 
@@ -97,12 +116,9 @@ define(["jquery", "footable", "humanize"],
                 item.action = "<div style='font-size:11px' class='label label-info'>" + item.action + "</div>";
             }
 
-            var score_content;
-            if (item.score < item.required_score) {
-                score_content = "<span class='text-success'>" + item.score.toFixed(2) + " / " + item.required_score + "</span>";
-            } else {
-                score_content = "<span class='text-danger'>" + item.score.toFixed(2) + " / " + item.required_score + "</span>";
-            }
+            var score_content = (item.score < item.required_score)
+                ? "<span class='text-success'>" + item.score.toFixed(2) + " / " + item.required_score + "</span>"
+                : "<span class='text-danger'>" + item.score.toFixed(2) + " / " + item.required_score + "</span>";
 
             item.score = {
                 options: {
@@ -112,22 +128,43 @@ define(["jquery", "footable", "humanize"],
             };
         }
 
+        function getSelector(id) {
+            var e = document.getElementById(id);
+            return e.options[e.selectedIndex].value;
+        }
+
+        function get_compare_function() {
+            var compare_functions = {
+                magnitude: function (e1, e2) {
+                    return Math.abs(e2.score) - Math.abs(e1.score);
+                },
+                name: function (e1, e2) {
+                    return e1.name.localeCompare(e2.name);
+                },
+                score: function (e1, e2) {
+                    return e2.score - e1.score;
+                }
+            };
+
+            return compare_functions[getSelector("selSymOrder")];
+        }
+
+        function sort_symbols(o, compare_function) {
+            return Object.keys(o)
+                .map(function (key) {
+                    return o[key];
+                })
+                .sort(compare_function)
+                .map(function (e) { return e.str; })
+                .join("<br>\n");
+        }
+
         function process_history_v2(data) {
-        // Display no more than rcpt_lim recipients
+            // Display no more than rcpt_lim recipients
             var rcpt_lim = 3;
             var items = [];
-
-            function getSelector(id) {
-                var e = document.getElementById(id);
-                return e.options[e.selectedIndex].value;
-            }
-            var compare = (getSelector("selSymOrder") === "score")
-                ? function (e1, e2) {
-                    return Math.abs(e1.score) < Math.abs(e2.score);
-                }
-                : function (e1, e2) {
-                    return e1.name.localeCompare(e2.name);
-                };
+            var unsorted_symbols = [];
+            var compare_function = get_compare_function();
 
             $("#selSymOrder, label[for='selSymOrder']").show();
 
@@ -152,43 +189,49 @@ define(["jquery", "footable", "humanize"],
                             full += item.rcpt_mime.join(", ");
                             shrt += item.rcpt_mime.slice(0, rcpt_lim).join(",&#8203;") + more("rcpt_mime");
                         }
-                        return {full: full, shrt: shrt};
+                        return {full:full, shrt:shrt};
+                    }
+
+                    function get_symbol_class(name, score) {
+                        if (name.match(/^GREYLIST$/)) {
+                            return "symbol-special";
+                        }
+
+                        if (score < 0) {
+                            return "symbol-negative";
+                        } else if (score > 0) {
+                            return "symbol-positive";
+                        }
+                        return null;
                     }
 
                     preprocess_item(item);
-                    Object.keys(item.symbols).map(function (key) {
-                        var str;
+                    Object.keys(item.symbols).forEach(function (key) {
                         var sym = item.symbols[key];
+                        sym.str = '<span class="symbol-default ' + get_symbol_class(sym.name, sym.score) + '"><strong>';
 
                         if (sym.description) {
-                            str = "<strong><abbr data-sym-key=\"" + key + "\">" + sym.name + "</abbr></strong>(" + sym.score + ")";
-
+                            sym.str += '<abbr data-sym-key="' + key + '">' +
+                                sym.name + "</abbr></strong> (" + sym.score + ")</span>";
                             // Store description for tooltip
                             symbolDescriptions[key] = sym.description;
                         } else {
-                            str = "<strong>" + sym.name + "</strong>(" + sym.score + ")";
+                            sym.str += sym.name + "</strong> (" + sym.score + ")</span>";
                         }
 
                         if (sym.options) {
-                            str += "[" + sym.options.join(",") + "]";
+                            sym.str += " [" + sym.options.join(",") + "]";
                         }
-                        item.symbols[key].str = str;
                     });
-                    item.symbols = Object.keys(item.symbols)
-                        .map(function (key) {
-                            return item.symbols[key];
-                        })
-                        .sort(compare)
-                        .map(function (e) { return e.str; })
-                        .join("<br>\n");
+                    unsorted_symbols.push(item.symbols);
+                    item.symbols = sort_symbols(item.symbols, compare_function);
                     item.time = {
                         value: unix_time_format(item.unix_time),
                         options: {
                             sortValue: item.unix_time
                         }
                     };
-                    var scan_time = item.time_real.toFixed(3) + " / " +
-                item.time_virtual.toFixed(3);
+                    var scan_time = item.time_real.toFixed(3) + " / " + item.time_virtual.toFixed(3);
                     item.scan_time = {
                         options: {
                             sortValue: item.time_real
@@ -214,7 +257,7 @@ define(["jquery", "footable", "humanize"],
                     items.push(item);
                 });
 
-            return items;
+            return {items:items, symbols:unsorted_symbols};
         }
 
         function process_history_legacy(data) {
@@ -252,7 +295,7 @@ define(["jquery", "footable", "humanize"],
                 items.push(item);
             });
 
-            return items;
+            return {items:items};
         }
 
         function columns_v2() {
@@ -329,7 +372,13 @@ define(["jquery", "footable", "humanize"],
                 sortValue: function (val) { return Number(val.options.sortValue); }
             }, {
                 name: "symbols",
-                title: "Symbols",
+                title: "Symbols<br /><br />" +
+                        '<span style="font-weight:normal;">Sort by:</span><br />' +
+                        '<div class="btn-group btn-group-xs btn-sym-order" data-toggle="buttons">' +
+                            '<button type="button" class="btn btn-default btn-sym-magnitude" value="magnitude">Magnitude</button>' +
+                            '<button type="button" class="btn btn-default btn-sym-score" value="score">Value</button>' +
+                            '<button type="button" class="btn btn-default btn-sym-name" value="name">Name</button>' +
+                        "</div>",
                 breakpoints: "all",
                 style: {
                     "font-size": "11px",
@@ -347,7 +396,7 @@ define(["jquery", "footable", "humanize"],
                 formatter: Humanize.compactInteger
             }, {
                 name: "scan_time",
-                title: "Scan time",
+                title: '<span title="real / virtual">Scan time</span>',
                 breakpoints: "xs sm md",
                 style: {
                     "font-size": "11px",
@@ -435,7 +484,7 @@ define(["jquery", "footable", "humanize"],
                 formatter: Humanize.compactInteger
             }, {
                 name: "scan_time",
-                title: "Scan time",
+                title: '<span title="real / virtual">Scan time</span>',
                 breakpoints: "xs sm",
                 style: {
                     "font-size": "11px",
@@ -499,38 +548,50 @@ define(["jquery", "footable", "humanize"],
             return func();
         }
 
-        ui.getHistory = function (rspamd, tables, neighbours, checked_server) {
+        function drawTooltips() {
+            // Update symbol description tooltips
+            $.each(symbolDescriptions, function (key, description) {
+                $("abbr[data-sym-key=" + key + "]").tooltip({
+                    placement: "bottom",
+                    html: true,
+                    title: description
+                });
+            });
+        }
+
+        function initHistoryTable(rspamd, tables, data, items) {
+            /* eslint-disable consistent-this, no-underscore-dangle, one-var-declaration-per-line */
             FooTable.actionFilter = FooTable.Filtering.extend({
-                construct : function (instance) {
+                construct: function (instance) {
                     this._super(instance);
                     this.actions = ["reject", "add header", "greylist",
                         "no action", "soft reject", "rewrite subject"];
                     this.def = "Any action";
                     this.$action = null;
                 },
-                $create : function () {
+                $create: function () {
                     this._super();
                     var self = this, $form_grp = $("<div/>", {
-                        class : "form-group"
+                        class: "form-group"
                     }).append($("<label/>", {
-                        class : "sr-only",
-                        text : "Action"
+                        class: "sr-only",
+                        text: "Action"
                     })).prependTo(self.$form);
 
                     self.$action = $("<select/>", {
-                        class : "form-control"
+                        class: "form-control"
                     }).on("change", {
-                        self : self
+                        self: self
                     }, self._onStatusDropdownChanged).append(
                         $("<option/>", {
-                            text : self.def
+                            text: self.def
                         })).appendTo($form_grp);
 
                     $.each(self.actions, function (i, action) {
                         self.$action.append($("<option/>").text(action));
                     });
                 },
-                _onStatusDropdownChanged : function (e) {
+                _onStatusDropdownChanged: function (e) {
                     var self = e.data.self, selected = $(this).val();
                     if (selected !== self.def) {
                         if (selected === "reject") {
@@ -543,7 +604,7 @@ define(["jquery", "footable", "humanize"],
                     }
                     self.filter();
                 },
-                draw : function () {
+                draw: function () {
                     this._super();
                     var action = this.find("action");
                     if (action instanceof FooTable.Filter) {
@@ -557,130 +618,152 @@ define(["jquery", "footable", "humanize"],
                     }
                 }
             });
+            /* eslint-enable consistent-this, no-underscore-dangle, one-var-declaration-per-line */
 
-            var drawTooltips = function () {
-            // Update symbol description tooltips
-                $.each(symbolDescriptions, function (key, description) {
-                    $("abbr[data-sym-key=" + key + "]").tooltip({
-                        placement: "bottom",
-                        html: true,
-                        title: description
-                    });
-                });
-            };
-
-            if (checked_server === "All SERVERS") {
-                rspamd.query("history", {
-                    success: function (req_data) {
-                        function differentVersions(neighbours_data) {
-                            var dv = neighbours_data.some(function (e) {
-                                return e.version !== neighbours_data[0].version;
-                            });
-                            if (dv) {
-                                rspamd.alertMessage("alert-error",
-                                    "Neighbours history backend versions do not match. Cannot display history.");
-                                return true;
-                            }
-                        }
-
-                        var neighbours_data = req_data
-                            .filter(function (d) { return d.status; }) // filter out unavailable neighbours
-                            .map(function (d) { return d.data; });
-                        if (neighbours_data.length && !differentVersions(neighbours_data)) {
-                            var data = {};
-                            if (neighbours_data[0].version) {
-                                data.rows = [].concat.apply([], neighbours_data
-                                    .map(function (e) {
-                                        return e.rows;
-                                    }));
-                                data.version = neighbours_data[0].version;
-                            } else {
-                            // Legacy version
-                                data = [].concat.apply([], neighbours_data);
-                            }
-
-                            var items = process_history_data(data);
-                            ft.history = FooTable.init("#historyTable", {
-                                columns: get_history_columns(data),
-                                rows: items,
-                                paging: {
-                                    enabled: true,
-                                    limit: 5,
-                                    size: 25
-                                },
-                                filtering: {
-                                    enabled: true,
-                                    position: "left",
-                                    connectors: false
-                                },
-                                sorting: {
-                                    enabled: true
-                                },
-                                components: {
-                                    filtering: FooTable.actionFilter
-                                },
-                                on: {
-                                    "ready.ft.table": drawTooltips,
-                                    "after.ft.sorting": drawTooltips,
-                                    "after.ft.paging": drawTooltips,
-                                    "after.ft.filtering": drawTooltips
-                                }
-                            });
-                        } else if (ft.history) {
-                            ft.history.destroy();
-                            delete ft.history;
-                        }
+            tables.history = FooTable.init("#historyTable", {
+                columns: get_history_columns(data),
+                rows: items,
+                paging: {
+                    enabled: true,
+                    limit: 5,
+                    size: page_size.history
+                },
+                filtering: {
+                    enabled: true,
+                    position: "left",
+                    connectors: false
+                },
+                sorting: {
+                    enabled: true
+                },
+                components: {
+                    filtering: FooTable.actionFilter
+                },
+                on: {
+                    "ready.ft.table": drawTooltips,
+                    "after.ft.sorting": drawTooltips,
+                    "after.ft.paging": drawTooltips,
+                    "after.ft.filtering": drawTooltips,
+                    "expand.ft.row": function (e, ft, row) {
+                        setTimeout(function () {
+                            var detail_row = row.$el.next();
+                            var order = getSelector("selSymOrder");
+                            detail_row.find(".btn-sym-" + order)
+                                .addClass("active").siblings().removeClass("active");
+                        }, 5);
                     }
-                });
-            } else {
-                $.ajax({
-                    dataType: "json",
-                    url: neighbours[checked_server].url + "history",
-                    jsonp: false,
-                    beforeSend: function (xhr) {
-                        xhr.setRequestHeader("Password", rspamd.getPassword());
-                    },
-                    error: function () {
-                        rspamd.alertMessage("alert-error", "Cannot receive history");
-                    },
-                    success: function (data) {
-                        var items = process_history_data(data);
-                        ft.history = FooTable.init("#historyTable", {
-                            columns: get_history_columns(data),
-                            rows: items,
-                            paging: {
-                                enabled: true,
-                                limit: 5,
-                                size: 25
-                            },
-                            filtering: {
-                                enabled: true,
-                                position: "left",
-                                connectors: false
-                            },
-                            sorting: {
-                                enabled: true
-                            },
-                            components: {
-                                filtering: FooTable.actionFilter
-                            },
-                            on: {
-                                "ready.ft.table": drawTooltips,
-                                "after.ft.sorting": drawTooltips,
-                                "after.ft.paging": drawTooltips,
-                                "after.ft.filtering": drawTooltips
-                            }
-                        });
-                    }
-                });
+                }
+            });
+        }
+
+        function destroyTable(tables, table) {
+            if (tables[table]) {
+                tables[table].destroy();
+                delete tables[table];
             }
+        }
+
+        ui.getHistory = function (rspamd, tables) {
+            function waitForRowsDisplayed(rows_total, callback, iteration) {
+                var i = (typeof iteration === "undefined") ? 10 : iteration;
+                var num_rows = $("#historyTable > tbody > tr").length;
+                if (num_rows === page_size.history ||
+                    num_rows === rows_total) {
+                    return callback();
+                } else if (--i) {
+                    setTimeout(function () {
+                        waitForRowsDisplayed(rows_total, callback, i);
+                    }, 500);
+                }
+                return null;
+            }
+
+            rspamd.query("history", {
+                success: function (req_data) {
+                    function differentVersions(neighbours_data) {
+                        var dv = neighbours_data.some(function (e) {
+                            return e.version !== neighbours_data[0].version;
+                        });
+                        if (dv) {
+                            rspamd.alertMessage("alert-error",
+                                "Neighbours history backend versions do not match. Cannot display history.");
+                            return true;
+                        }
+                        return false;
+                    }
+
+                    var neighbours_data = req_data
+                        .filter(function (d) { return d.status; }) // filter out unavailable neighbours
+                        .map(function (d) { return d.data; });
+                    if (neighbours_data.length && !differentVersions(neighbours_data)) {
+                        var data = {};
+                        var version = neighbours_data[0].version;
+                        if (version) {
+                            data.rows = [].concat.apply([], neighbours_data
+                                .map(function (e) {
+                                    return e.rows;
+                                }));
+                            data.version = version;
+                        } else {
+                            // Legacy version
+                            data = [].concat.apply([], neighbours_data);
+                        }
+                        var o = process_history_data(data);
+                        var items = o.items;
+                        symbols = o.symbols;
+
+                        if (Object.prototype.hasOwnProperty.call(tables, "history") &&
+                            version === prevVersion) {
+                            tables.history.rows.load(items);
+                            if (version) { // Non-legacy
+                                // Is there a way to get an event when all rows are loaded?
+                                waitForRowsDisplayed(items.length, function () {
+                                    drawTooltips();
+                                });
+                            }
+                        } else {
+                            destroyTable(tables, "history");
+                            // Is there a way to get an event when the table is destroyed?
+                            setTimeout(function () {
+                                initHistoryTable(rspamd, tables, data, items);
+                            }, 200);
+                        }
+                        prevVersion = version;
+                    } else {
+                        destroyTable(tables, "history");
+                    }
+                },
+                errorMessage: "Cannot receive history",
+            });
+        };
+
+        ui.setup = function (rspamd, tables) {
+            function change_symbols_order(order) {
+                $(".btn-sym-" + order).addClass("active").siblings().removeClass("active");
+                var compare_function = get_compare_function();
+                $.each(tables.history.rows.all, function (i, row) {
+                    var cell_val = sort_symbols(symbols[i], compare_function);
+                    row.cells[8].val(cell_val, false, true);
+                });
+                drawTooltips();
+            }
+
             $("#updateHistory").off("click");
             $("#updateHistory").on("click", function (e) {
                 e.preventDefault();
-                ui.getHistory(rspamd, tables, neighbours, checked_server);
+                ui.getHistory(rspamd, tables);
             });
             $("#selSymOrder").unbind().change(function () {
-                ui.getHistory(rspamd, tables, neighbours, checked_server);
+                var order = this.value;
+                change_symbols_order(order);
+            });
+            $("#history_page_size").change(function () {
+                set_page_size(this.value, function (n) { tables.history.pageSize(n); });
+            });
+            $(document).on("click", ".btn-sym-order button", function () {
+                var order = this.value;
+                $("#selSymOrder").val(order);
+                change_symbols_order(order);
             });
 
             // @reset history log
@@ -690,46 +773,34 @@ define(["jquery", "footable", "humanize"],
                 if (!confirm("Are you sure you want to reset history log?")) { // eslint-disable-line no-alert
                     return;
                 }
-                if (ft.history) {
-                    ft.history.destroy();
-                    delete ft.history;
-                }
-                if (ft.errors) {
-                    ft.errors.destroy();
-                    delete ft.errors;
-                }
+                destroyTable(tables, "history");
+                destroyTable(tables, "errors");
 
                 rspamd.query("historyreset", {
                     success: function () {
-                        ui.getHistory(rspamd, tables, neighbours, checked_server);
-                        ui.getErrors(rspamd, tables, neighbours, checked_server);
+                        ui.getHistory(rspamd, tables);
+                        ui.getErrors(rspamd, tables);
                     },
                     errorMessage: "Cannot reset history log"
                 });
             });
         };
 
-        function drawErrorsTable(data) {
-            var items = [];
-            $.each(data, function (i, item) {
-                items.push(
-                    item.ts = unix_time_format(item.ts)
-                );
-            });
-            ft.errors = FooTable.init("#errorsLog", {
+        function initErrorsTable(tables, rows) {
+            tables.errors = FooTable.init("#errorsLog", {
                 columns: [
-                    {sorted: true, direction: "DESC", name:"ts", title:"Time", style:{"font-size":"11px", "width":300, "maxWidth":300}},
+                    {sorted:true, direction:"DESC", name:"ts", title:"Time", style:{"font-size":"11px", "width":300, "maxWidth":300}},
                     {name:"type", title:"Worker type", breakpoints:"xs sm", style:{"font-size":"11px", "width":150, "maxWidth":150}},
                     {name:"pid", title:"PID", breakpoints:"xs sm", style:{"font-size":"11px", "width":110, "maxWidth":110}},
                     {name:"module", title:"Module", style:{"font-size":"11px"}},
                     {name:"id", title:"Internal ID", style:{"font-size":"11px"}},
                     {name:"message", title:"Message", breakpoints:"xs sm", style:{"font-size":"11px"}},
                 ],
-                rows: data,
+                rows: rows,
                 paging: {
                     enabled: true,
                     limit: 5,
-                    size: 25
+                    size: page_size.errors
                 },
                 filtering: {
                     enabled: true,
@@ -742,40 +813,34 @@ define(["jquery", "footable", "humanize"],
             });
         }
 
-        ui.getErrors = function (rspamd, tables, neighbours, checked_server) {
+        ui.getErrors = function (rspamd, tables) {
             if (rspamd.read_only) return;
 
-            if (checked_server !== "All SERVERS") {
-                $.ajax({
-                    dataType: "json",
-                    url: neighbours[checked_server].url + "errors",
-                    jsonp: false,
-                    beforeSend: function (xhr) {
-                        xhr.setRequestHeader("Password", rspamd.getPassword());
-                    },
-                    success: function (data) {
-                        drawErrorsTable(data);
-                    },
-                    errorMessage: "Cannot receive errors"
-                });
-            } else {
-                rspamd.query("errors", {
-                    success: function (req_data) {
-                        var neighbours_data = req_data
-                            .filter(function (d) {
-                                return d.status;
-                            }) // filter out unavailable neighbours
-                            .map(function (d) {
-                                return d.data;
-                            });
-                        drawErrorsTable([].concat.apply([], neighbours_data));
+            rspamd.query("errors", {
+                success: function (data) {
+                    var neighbours_data = data
+                        .filter(function (d) {
+                            return d.status;
+                        }) // filter out unavailable neighbours
+                        .map(function (d) {
+                            return d.data;
+                        });
+                    var rows = [].concat.apply([], neighbours_data);
+                    $.each(rows, function (i, item) {
+                        item.ts = unix_time_format(item.ts);
+                    });
+                    if (Object.prototype.hasOwnProperty.call(tables, "errors")) {
+                        tables.errors.rows.load(rows);
+                    } else {
+                        initErrorsTable(tables, rows);
                     }
-                });
-            }
+                }
+            });
+
             $("#updateErrors").off("click");
             $("#updateErrors").on("click", function (e) {
                 e.preventDefault();
-                ui.getErrors(rspamd, tables, neighbours, checked_server);
+                ui.getErrors(rspamd, tables);
             });
         };
 
